#!/usr/bin/env python3

import argparse
import json
import os
import re
import subprocess  # nosec

import multi_repo_automation as mra
import ruamel.yaml
import security_md


def _get_default_branch() -> str:
    """Get the default branch name."""
    return subprocess.run(
        ["gh", "repo", "view", "--json", "defaultBranchRef", "--jq", ".defaultBranchRef.name"],
        stdout=subprocess.PIPE,
        encoding="utf-8",
        check=True,
    ).stdout.strip()


def main() -> None:
    """Create a new version with its stabilization branch."""
    args_parser = argparse.ArgumentParser(
        description="Create a new version with its stabilization branch",
        usage="""
This will:
- Stash all your changes
- Checkout the master branch
- Pull it from origin
- Push it to a new stabilization branch
- Checkout a new branch named new-version
- Do the changes for the new version
  - Update the SECURITY.md config
  - Update the Renovate config
  - Update the audit workflow
  - Create the backport label
- Push it
- Create a pull request
- Go back to your old branch

If you run the tool without any version it will check that everything is OK
regarding the SECURITY.md available on GitHub.
    """,
    )
    args_parser.add_argument(
        "--version",
        help="The version to create",
    )
    args_parser.add_argument(
        "--force",
        action="store_true",
        help="Force create the branch and push it",
    )
    args_parser.add_argument(
        "--supported-until",
        help="The date until the version is supported, can also be To be defined or Best effort",
        default="Best effort",
    )
    args_parser.add_argument(
        "--upstream-supported-until",
        help="The date until the version is supported upstream",
    )
    args_parser.add_argument(
        "--remote",
        default="origin",
        help="The remote to push the branch",
    )

    arguments = args_parser.parse_args()

    # Get the repo information e.g.:
    # {
    #     "name": "camptocamp/tag_publish",
    #     "remote": "origin",
    #     "dir": "/home/user/src/tag_publish",
    # }
    # can be override with a repo.yaml file
    repo = mra.get_repo_config()

    # Stash all your changes
    subprocess.run(["git", "stash", "--all", "--message=Stashed by release creation"], check=True)
    old_branch_name = subprocess.run(
        ["git", "rev-parse", "--abbrev-ref", "HEAD"],
        stdout=subprocess.PIPE,
        check=True,
    ).stdout.strip()

    default_branch = _get_default_branch()

    # Checkout the master branch
    subprocess.run(["git", "checkout", default_branch], check=True)

    # Pull it from origin
    subprocess.run(["git", "pull", arguments, default_branch], check=True)

    # Push it to a new stabilization branch
    if arguments.version:
        subprocess.run(
            [
                "git",
                "push",
                *(["--force"] if arguments.force else []),
                repo.get("remote", "origin"),
                f"HEAD:{arguments.version}",
            ],
            check=not arguments.force,
        )

    version = arguments.version
    branch_name = "new-version" if version is None else f"new-version-{version}"

    # Checkout a new branch named new-version
    if arguments.force:
        subprocess.run(["git", "branch", "-D", branch_name])  # pylint: disable=subprocess-run-check
    subprocess.run(["git", "checkout", "-b", branch_name], check=True)

    # # # Do the changes for the new version # # #

    remotes = [
        r
        for r in subprocess.run(
            ["git", "remote"], stdout=subprocess.PIPE, encoding="utf-8", check=True
        ).stdout.split()
        if r != ""
    ]
    remote_branches = [
        b.strip()[len("remotes/") :]
        for b in subprocess.run(
            ["git", "branch", "--all"], stdout=subprocess.PIPE, encoding="utf-8", check=True
        ).stdout.split()
        if b != "" and b.strip().startswith("remotes/")
    ]
    if "upstream" in remotes:
        remote_branches = [b[len("upstream") + 1 :] for b in remote_branches if b.startswith("upstream/")]
    elif "origin" in remotes:
        remote_branches = [b[len("origin") + 1 :] for b in remote_branches if b.startswith("origin/")]
    else:
        remote_branches = ["/".join(b.split("/")[1:]) for b in remote_branches]

    if os.path.exists("SECURITY.md"):
        with open("SECURITY.md", encoding="utf-8") as security_file:
            security_text = security_file.read()
            security = security_md.Security(security_text)

        stabilization_branches = security.branches()
    else:
        stabilization_branches = []
    modified_files = []

    if version:
        stabilization_branches.append(version)

        if os.path.exists("SECURITY.md"):
            modified_files.append("SECURITY.md")
            with mra.Edit("SECURITY.md") as security_md_file:
                security_md_lines = security_md_file.data.split("\n")
                index = -1
                for i, line in enumerate(security_md_lines):
                    if line.startswith("| "):
                        index = i

                new_line = f"| {version} | {arguments.supported_until} |"
                if arguments.upstream_supported_until:
                    new_line += f" {arguments.upstream_supported_until} |"

                security_md_file.data = "\n".join(
                    [*security_md_lines[: index + 1], new_line, *security_md_lines[index + 1 :]]
                )

    stabilization_branches_with_master = [*stabilization_branches, default_branch]

    for labels in mra.gh_json("label", ["name"], "list"):
        if (
            labels["name"].startswith("backport ")
            and labels["name"].replace("backport ", "") not in stabilization_branches_with_master
        ):
            mra.gh("label", "delete", labels["name"], "--yes")

    for branch in stabilization_branches_with_master:
        mra.gh(
            "label",
            "create",
            "--force",
            f"backport {branch}",
            "--color=5aed94",
            f"--description=Backport the pull request to the '{branch}' branch",
        )

    if os.path.exists(".github/renovate.json5"):
        modified_files.append(".github/renovate.json5")
        with mra.EditRenovateConfig(".github/renovate.json5") as renovate_config:
            if stabilization_branches:
                if "baseBranches: " in renovate_config.data:
                    renovate_config.data = re.sub(
                        r"(.*baseBranches: )\[[^\]]*\](.*)",
                        rf"\1{json.dumps(stabilization_branches_with_master)}\2",
                        renovate_config.data,
                    )
                else:
                    renovate_config.add(
                        f"baseBranches: {json.dumps(stabilization_branches_with_master)},\n", "baseBranches"
                    )

    if stabilization_branches and os.path.exists(".github/workflows/audit.yaml"):
        modified_files.append(".github/workflows/audit.yaml")
        with mra.EditYAML(".github/workflows/audit.yaml") as yaml:
            for job in yaml["jobs"].values():
                matrix = job.get("strategy", {}).get("matrix", {})
                if "include" in matrix and version:
                    new_include = dict(matrix["include"][-1])
                    new_include["branch"] = version
                    matrix["include"].append(new_include)

                if "branch" in matrix and stabilization_branches:
                    yaml_stabilization_branches = ruamel.yaml.comments.CommentedSeq(stabilization_branches)
                    yaml_stabilization_branches._yaml_add_comment(  # pylint: disable=protected-access
                        [
                            ruamel.yaml.CommentToken("\n\n", ruamel.yaml.error.CommentMark(0), None),
                            None,
                            None,
                            None,
                        ],
                        len(stabilization_branches) - 1,
                    )
                    job["strategy"]["matrix"]["branch"] = yaml_stabilization_branches

    # Commit the changes
    message = f"Create the new version '{version}'" if version else "Update the supported versions"
    if os.path.exists(".pre-commit-config.yaml"):
        subprocess.run(["pre-commit", "run", "--color=never", "--all-files"], check=False)
    subprocess.run(["git", "add", *modified_files], check=True)
    subprocess.run(["git", "commit", f"--message={message}"], check=True)

    # Push it
    subprocess.run(
        [
            "git",
            "push",
            *(["--force"] if arguments.force else []),
            repo.get("remote", "origin"),
            branch_name,
        ],
        check=True,
    )

    # Create a pull request
    url = mra.gh(
        "pr",
        "create",
        f"--title={message}",
        "--body=",
        f"--head={branch_name}",
        f"--base={repo.get('master_branch', 'master')}",
    ).strip()

    # Go back to your old branch
    subprocess.run(["git", "checkout", old_branch_name, "--"], check=True)

    if url:
        pr_number = url.rsplit("/", 1)[-1]
        mra.gh("browse", pr_number)
    else:
        mra.gh("browse")


if __name__ == "__main__":
    main()



#!/usr/bin/env python3

"""
The publish script.
"""

import argparse
import json
import os
import os.path
import random
import re
import subprocess  # nosec
import sys
from re import Match
from typing import Optional, cast

import security_md
import yaml

import tag_publish
import tag_publish.configuration
import tag_publish.lib.docker
import tag_publish.lib.oidc
import tag_publish.publish


def match(tpe: str, base_re: str) -> Optional[Match[str]]:
    """
    Return the match for `GITHUB_REF` basically like: `refs/<tpe>/<base_re>`.

    Arguments:
        tpe: The type of ref we want to match (heads, tag, ...)
        base_re: The regular expression to match the value

    """
    if base_re[0] == "^":
        base_re = base_re[1:]
    if base_re[-1] != "$":
        base_re += "$"
    return re.match(f"^refs/{tpe}/{base_re}", os.environ["GITHUB_REF"])


def to_version(full_config: tag_publish.configuration.Configuration, value: str, kind: str) -> str:
    """
    Compute publish version from branch name or tag.

    Arguments:
        full_config: The full configuration
        value: The value to be transformed
        kind: The name of the transformer in the configuration

    """
    item_re = tag_publish.compile_re(
        cast(
            tag_publish.configuration.VersionTransform,
            full_config["version"].get(kind + "_to_version_re", []),
        )
    )
    value_match = tag_publish.match(value, item_re)
    if value_match[0] is not None:
        return tag_publish.get_value(*value_match)
    return value


def main() -> None:
    """
    Run the publish.
    """
    parser = argparse.ArgumentParser(description="Publish the project.")
    parser.add_argument("--group", default="default", help="The publishing group")
    parser.add_argument("--version", help="The version to publish to")
    parser.add_argument(
        "--docker-versions",
        help="The versions to publish on Docker registry, comma separated, ex: 'x,x.y,x.y.z,latest'.",
    )
    parser.add_argument("--snyk-version", help="The version to publish to Snyk")
    parser.add_argument("--branch", help="The branch from which to compute the version")
    parser.add_argument("--tag", help="The tag from which to compute the version")
    parser.add_argument("--dry-run", action="store_true", help="Don't do the publish")
    parser.add_argument("--dry-run-tag", help="Don't do the publish, on a tag")
    parser.add_argument("--dry-run-branch", help="Don't do the publish, on a branch")
    parser.add_argument(
        "--type",
        help="The type of version, if no argument provided auto-determinate, can be: "
        "rebuild (in case of rebuild), version_tag, version_branch, feature_branch, feature_tag "
        "(for pull request)",
    )
    args = parser.parse_args()

    if args.dry_run_tag is not None:
        args.dry_run = True
        os.environ["GITHUB_REF"] = f"refs/tags/{args.dry_run_tag}"
    if args.dry_run_branch is not None:
        args.dry_run = True
        os.environ["GITHUB_REF"] = f"refs/heads/{args.dry_run_branch}"

    github = tag_publish.GH()
    config = tag_publish.get_config(github)

    # Describe the kind of release we do: rebuild (specified with --type), version_tag, version_branch,
    # feature_branch, feature_tag (for pull request)
    version: str = ""
    ref = os.environ.get("GITHUB_REF", "refs/heads/fake-local")
    local = "GITHUB_REF" not in os.environ

    if len([e for e in [args.version, args.branch, args.tag] if e is not None]) > 1:
        print("::error::you specified more than one of the arguments --version, --branch or --tag")
        sys.exit(1)

    version_type = args.type

    tag_match = tag_publish.match(
        ref,
        tag_publish.compile_re(config["version"].get("tag_to_version_re", []), "refs/tags/"),
    )
    branch_match = tag_publish.match(
        ref,
        tag_publish.compile_re(config["version"].get("branch_to_version_re", []), "refs/heads/"),
    )
    ref_match = re.match(r"refs/pull/(.*)/merge", ref)

    if args.version is not None:
        version = args.version
    elif args.branch is not None:
        version = to_version(config, args.branch, "branch")
    elif args.tag is not None:
        version = to_version(config, args.tag, "tag")
    elif tag_match[0] is not None:
        if version_type is None:
            version_type = "version_tag"
        else:
            print("::warning::you specified the argument --type but not one of --version, --branch or --tag")
        version = tag_publish.get_value(*tag_match)
    elif branch_match[0] is not None:
        if version_type is None:
            version_type = "version_branch"
        else:
            print("::warning::you specified the argument --type but not one of --version, --branch or --tag")
        version = tag_publish.get_value(*branch_match)
    elif ref_match is not None:
        version = tag_publish.get_value(ref_match, {}, ref)
        if version_type is None:
            version_type = "feature_branch"
    elif ref.startswith("refs/heads/"):
        if version_type is None:
            version_type = "feature_branch"
        else:
            print("::warning::you specified the argument --type but not one of --version, --branch or --tag")
        # By the way we replace '/' by '_' because it isn't supported by Docker
        version = "_".join(ref.split("/")[2:])
    elif ref.startswith("refs/tags/"):
        if version_type is None:
            version_type = "feature_tag"
        else:
            print("::warning::you specified the argument --type but not one of --version, --branch or --tag")
        # By the way we replace '/' by '_' because it isn't supported by Docker
        version = "_".join(ref.split("/")[2:])
    else:
        print(
            f"WARNING: {ref} is not supported, only ref starting with 'refs/heads/' or 'refs/tags/' "
            "are supported, ignoring"
        )
        sys.exit(0)

    if version_type is None:
        print(
            "::error::you specified one of the arguments --version, --branch or --tag but not the --type, "
            f"GitHub ref is: {ref}"
        )
        sys.exit(1)

    if version_type is not None:
        if args.dry_run:
            print(f"Create release type {version_type}: {version} (dry run)")
        else:
            print(f"Create release type {version_type}: {version}")

    github = tag_publish.GH()

    success = True
    published_payload: list[tag_publish.PublishedPayload] = []

    success &= _handle_pypi_publish(
        args.group, args.dry_run, config, version, version_type, github, published_payload
    )
    success &= _handle_docker_publish(
        args.group,
        args.dry_run,
        args.docker_versions,
        args.snyk_version,
        config,
        version,
        version_type,
        github,
        published_payload,
        local,
    )
    success &= _handle_helm_publish(args.dry_run, config, version, version_type, github, published_payload)
    _trigger_dispatch_events(config, published_payload, github)

    if not success:
        sys.exit(1)


def _handle_pypi_publish(
    group: str,
    dry_run: bool,
    config: tag_publish.configuration.Configuration,
    version: str,
    version_type: str,
    github: tag_publish.GH,
    published_payload: list[tag_publish.PublishedPayload],
) -> bool:
    success = True
    pypi_config = config.get("pypi", {})
    if pypi_config:
        if "packages" in pypi_config:
            tag_publish.lib.oidc.pypi_login()

        for package in pypi_config.get("packages", []):
            if package.get("group", tag_publish.configuration.PIP_PACKAGE_GROUP_DEFAULT) == group:
                publish = version_type in pypi_config.get(
                    "versions", tag_publish.configuration.PYPI_VERSIONS_DEFAULT
                )
                folder = package.get("folder", tag_publish.configuration.PYPI_PACKAGE_FOLDER_DEFAULT)
                if dry_run:
                    print(f"{'Publishing' if publish else 'Checking'} '{folder}' to pypi, skipping (dry run)")
                else:
                    success &= tag_publish.publish.pip(package, version, version_type, publish, github)
                    published_payload.append(
                        {
                            "type": "pypi",
                            "folder": folder,
                            "version": version,
                            "version_type": version_type,
                        }
                    )
    return success


def _handle_docker_publish(
    group: str,
    dry_run: bool,
    docker_versions: str,
    snyk_version: str,
    config: tag_publish.configuration.Configuration,
    version: str,
    version_type: str,
    github: tag_publish.GH,
    published_payload: list[tag_publish.PublishedPayload],
    local: bool,
) -> bool:
    success = True
    docker_config = config.get("docker", {})
    if docker_config:
        if docker_config.get("auto_login", tag_publish.configuration.DOCKER_AUTO_LOGIN_DEFAULT):
            subprocess.run(
                [
                    "docker",
                    "login",
                    "ghcr.io",
                    "--username=github",
                    f"--password={os.environ['GITHUB_TOKEN']}",
                ],
                check=True,
            )
        security_text = ""
        if local:
            with open("SECURITY.md", encoding="utf-8") as security_file:
                security_text = security_file.read()
                security = security_md.Security(security_text)
        else:
            security = tag_publish.get_security_md(github)

        version_index = security.version_index
        alternate_tag_index = security.alternate_tag_index

        row_index = -1
        if version_index >= 0:
            for index, row in enumerate(security.data):
                if row[version_index] == version:
                    row_index = index
                    break

        alt_tags = set()
        if alternate_tag_index >= 0 and row_index >= 0:
            alt_tags = {
                t.strip() for t in security.data[row_index][alternate_tag_index].split(",") if t.strip()
            }
        if version_index >= 0 and security.data[-1][version_index] == version:
            add_latest = True
            for data in security.data:
                row_tags = {t.strip() for t in data[alternate_tag_index].split(",") if t.strip()}
                if "latest" in row_tags:
                    print("latest found in ", row_tags)
                    add_latest = False
                    break
            if add_latest:
                alt_tags.add("latest")

        images_src: set[str] = set()
        images_full: list[str] = []
        images_snyk: set[str] = set()
        versions = docker_versions.split(",") if docker_versions else [version]
        for image_conf in docker_config.get("images", []):
            if image_conf.get("group", tag_publish.configuration.DOCKER_IMAGE_GROUP_DEFAULT) == group:
                for tag_config in image_conf.get("tags", tag_publish.configuration.DOCKER_IMAGE_TAGS_DEFAULT):
                    tag_src = tag_config.format(version="latest")
                    image_source = f"{image_conf['name']}:{tag_src}"
                    images_src.add(image_source)
                    tag_snyk = tag_config.format(version=snyk_version or version).lower()
                    image_snyk = f"{image_conf['name']}:{tag_snyk}"

                    # Workaround sine we have the business plan
                    image_snyk = f"{image_conf['name']}_{tag_snyk}"

                    if not dry_run:
                        subprocess.run(["docker", "tag", image_source, image_snyk], check=True)
                    images_snyk.add(image_snyk)
                    if tag_snyk != tag_src and not dry_run:
                        subprocess.run(
                            [
                                "docker",
                                "tag",
                                image_source,
                                f"{image_conf['name']}:{tag_snyk}",
                            ],
                            check=True,
                        )

                    for name, conf in docker_config.get(
                        "repository",
                        cast(
                            dict[str, tag_publish.configuration.DockerRepository],
                            tag_publish.configuration.DOCKER_REPOSITORY_DEFAULT,
                        ),
                    ).items():
                        for docker_version in versions:
                            if version_type in conf.get(
                                "versions",
                                tag_publish.configuration.DOCKER_REPOSITORY_VERSIONS_DEFAULT,
                            ):
                                tags = [
                                    tag_config.format(version=alt_tag)
                                    for alt_tag in [docker_version, *alt_tags]
                                ]

                                if dry_run:
                                    for tag in tags:
                                        print(
                                            f"Publishing {image_conf['name']}:{tag} to {name}, skipping "
                                            "(dry run)"
                                        )
                                else:
                                    success &= tag_publish.publish.docker(
                                        conf,
                                        name,
                                        image_conf,
                                        tag_src,
                                        tags,
                                        images_full,
                                        version_type,
                                        published_payload,
                                    )

        if dry_run:
            sys.exit(0)

        has_gopass = subprocess.run(["gopass", "--version"]).returncode == 0  # nosec # pylint: disable=subprocess-run-check
        if "SNYK_TOKEN" in os.environ or has_gopass:
            snyk_exec, env = tag_publish.snyk_exec()
            for image in images_snyk:
                print(f"::group::Snyk check {image}")
                sys.stdout.flush()
                sys.stderr.flush()
                try:
                    if version_type in ("version_branch", "version_tag"):
                        monitor_args = docker_config.get("snyk", {}).get(
                            "monitor_args",
                            tag_publish.configuration.DOCKER_SNYK_MONITOR_ARGS_DEFAULT,
                        )
                        if monitor_args is not False:
                            subprocess.run(  # pylint: disable=subprocess-run-check
                                [
                                    snyk_exec,
                                    "container",
                                    "monitor",
                                    *monitor_args,
                                    # Available only on the business plan
                                    # f"--project-tags=tag={image.split(':')[-1]}",
                                    image,
                                ],
                                env=env,
                            )
                    test_args = docker_config.get("snyk", {}).get(
                        "test_args", tag_publish.configuration.DOCKER_SNYK_TEST_ARGS_DEFAULT
                    )
                    snyk_error = False
                    if test_args is not False:
                        proc = subprocess.run(
                            [snyk_exec, "container", "test", *test_args, image],
                            check=False,
                            env=env,
                        )
                        if proc.returncode != 0:
                            snyk_error = True
                    print("::endgroup::")
                    if snyk_error:
                        print("::error::Critical vulnerability found by Snyk in the published image.")
                except subprocess.CalledProcessError as exception:
                    print(f"Error: {exception}")
                    print("::endgroup::")
                    print("::error::With error")

        versions_config, dpkg_config_found = tag_publish.lib.docker.get_versions_config()
        dpkg_success = True
        for image in images_src:
            dpkg_success &= tag_publish.lib.docker.check_versions(versions_config.get(image, {}), image)

        if not dpkg_success:
            current_versions_in_images: dict[str, dict[str, str]] = {}
            if dpkg_config_found:
                with open(".github/dpkg-versions.yaml", encoding="utf-8") as dpkg_versions_file:
                    current_versions_in_images = yaml.load(dpkg_versions_file, Loader=yaml.SafeLoader)
            for image in images_src:
                _, versions_image = tag_publish.lib.docker.get_dpkg_packages_versions(image)
                for dpkg_package, package_version in versions_image.items():
                    if dpkg_package not in current_versions_in_images.get(image, {}):
                        current_versions_in_images.setdefault(image, {})[dpkg_package] = str(package_version)
                for dpkg_package in list(current_versions_in_images[image].keys()):
                    if dpkg_package not in versions_image:
                        del current_versions_in_images[image][dpkg_package]
            if dpkg_config_found:
                print(
                    "::error::Some packages are have a greater version in the config raster then "
                    "in the image."
                )
            print("Current versions of the Debian packages in Docker images:")
            print(yaml.dump(current_versions_in_images, Dumper=yaml.SafeDumper, default_flow_style=False))
            if dpkg_config_found:
                with open(".github/dpkg-versions.yaml", "w", encoding="utf-8") as dpkg_versions_file:
                    yaml.dump(
                        current_versions_in_images,
                        dpkg_versions_file,
                        Dumper=yaml.SafeDumper,
                        default_flow_style=False,
                    )

            if dpkg_config_found:
                success = False
    return success


def _handle_helm_publish(
    dry_run: bool,
    config: tag_publish.configuration.Configuration,
    version: str,
    version_type: str,
    github: tag_publish.GH,
    published_payload: list[tag_publish.PublishedPayload],
) -> bool:
    success = True
    helm_config = config.get("helm", {})
    if helm_config.get("folders") and version_type in helm_config.get(
        "versions", tag_publish.configuration.HELM_VERSIONS_DEFAULT
    ):
        tag_publish.download_application("helm/chart-releaser")

        owner = github.repo.owner.login
        repo = github.repo.name
        commit_sha = (
            subprocess.run(["git", "rev-parse", "HEAD"], check=True, stdout=subprocess.PIPE)
            .stdout.strip()
            .decode()
        )
        if version_type == "version_branch":
            last_tag = (
                subprocess.run(
                    ["git", "describe", "--abbrev=0", "--tags"], check=True, stdout=subprocess.PIPE
                )
                .stdout.strip()
                .decode()
            )
            expression = re.compile(r"^[0-9]+\.[0-9]+\.[0-9]+$")
            while expression.match(last_tag) is None:
                last_tag = (
                    subprocess.run(
                        ["git", "describe", "--abbrev=0", "--tags", f"{last_tag}^"],
                        check=True,
                        stdout=subprocess.PIPE,
                    )
                    .stdout.strip()
                    .decode()
                )

            versions = last_tag.split(".")
            versions[-1] = str(int(versions[-1]) + 1)
            version = ".".join(versions)

        for folder in helm_config["folders"]:
            if dry_run:
                print(f"Publishing '{folder}' to helm, skipping (dry run)")
            else:
                token = os.environ["GITHUB_TOKEN"]
                success &= tag_publish.publish.helm(folder, version, owner, repo, commit_sha, token)
                published_payload.append(
                    {
                        "type": "helm",
                        "folder": folder,
                        "version": version,
                        "version_type": version_type,
                    }
                )
    return success


def _trigger_dispatch_events(
    config: tag_publish.configuration.Configuration,
    published_payload: list[tag_publish.PublishedPayload],
    github: tag_publish.GH,
) -> None:
    for published in published_payload:
        for dispatch_config in config.get("dispatch", []):
            repository = dispatch_config.get("repository")
            event_type = dispatch_config.get(
                "event-type", tag_publish.configuration.DISPATCH_EVENT_TYPE_DEFAULT
            )

            id_ = random.randint(1, 100000)  # nosec # noqa: S311
            published["id"] = id_

            if repository:
                print(f"Triggering {event_type}:{id_} on {repository} with {json.dumps(published)}")
                github_repo = github.github.get_repo(repository)
            else:
                print(f"Triggering {event_type}:{id_} with {json.dumps(published)}")
                github_repo = github.repo
            github_repo.create_repository_dispatch(event_type, published)  # type: ignore[arg-type]


if __name__ == "__main__":
    main()



"""
Tag Publish main module.
"""

import json
import os.path
import pkgutil
import re
import subprocess  # nosec
from re import Match, Pattern
from typing import Any, Optional, TypedDict, cast, overload

import applications_download
import github
import jsonschema_validator
import requests
import ruamel.yaml
import security_md
import yaml

import tag_publish.configuration

VersionTransform = TypedDict(
    "VersionTransform",
    {
        # The from regular expression
        "from": Pattern[str],
        # The expand regular expression: https://docs.python.org/3/library/re.html#re.Match.expand
        "to": str,
    },
    total=False,
)


class GH:
    """GitHub helper class."""

    def __init__(self) -> None:
        """Initialize the GitHub helper class."""
        token = (
            os.environ["GITHUB_TOKEN"]
            if "GITHUB_TOKEN" in os.environ
            else subprocess.run(
                ["gh", "auth", "token"], check=True, stdout=subprocess.PIPE, encoding="utf-8"
            ).stdout.strip()
        )
        self.auth = github.Auth.Token(token)
        self.github = github.Github(auth=self.auth)
        self.repo = self.github.get_repo(
            os.environ["GITHUB_REPOSITORY"]
            if "GITHUB_REPOSITORY" in os.environ
            else subprocess.run(
                ["gh", "repo", "view", "--json", "name,owner", "--jq", '(.owner.login + "/" + .name)'],
                check=True,
                stdout=subprocess.PIPE,
                encoding="utf-8",
            ).stdout.strip()
        )
        self.default_branch = self.repo.default_branch


def get_security_md(gh: GH) -> security_md.Security:
    """
    Get the SECURITY.md file.

    Arguments:
        gh: The GitHub helper

    """
    security_file = gh.repo.get_contents("SECURITY.md")
    assert isinstance(security_file, github.ContentFile.ContentFile)
    return security_md.Security(security_file.decoded_content.decode("utf-8"))


def merge(default_config: Any, config: Any) -> Any:
    """
    Deep merge the dictionaries (on dictionaries only, not on arrays).

    Arguments:
        default_config: The default config that will be applied
        config: The base config, will be modified

    """
    if not isinstance(default_config, dict) or not isinstance(config, dict):
        return config

    for key in default_config:
        if key not in config:
            config[key] = default_config[key]
        else:
            merge(default_config[key], config[key])
    return config


def get_config(gh: GH) -> tag_publish.configuration.Configuration:
    """
    Get the configuration, with project and auto detections.
    """
    config: tag_publish.configuration.Configuration = {}
    if os.path.exists(".github/publish.yaml"):
        schema_data = pkgutil.get_data("tag_publish", "schema.json")
        assert schema_data is not None
        schema = json.loads(schema_data)

        with open(".github/publish.yaml", encoding="utf-8") as open_file:
            yaml_ = ruamel.yaml.YAML()
            config = yaml_.load(open_file)
            jsonschema_validator.validate(".github/publish.yaml", cast(dict[str, Any], config), schema)

    merge(
        {
            "version": {
                "tag_to_version_re": [
                    {"from": r"([0-9]+.[0-9]+.[0-9]+)", "to": r"\1"},
                ],
                "branch_to_version_re": [
                    {"from": r"([0-9]+.[0-9]+)", "to": r"\1"},
                    {"from": gh.default_branch, "to": gh.default_branch},
                ],
            }
        },
        config,
    )

    return config


def get_value(matched: Optional[Match[str]], config: Optional[VersionTransform], value: str) -> str:
    """
    Get the final value.

    `match`, `config` and `value` are the result of `match`.

    The `config` should have a `to` key with an expand template.

    Arguments:
        matched: The matched object to a regular expression
        config: The result of `compile`
        value: The default value on returned no match

    Return the value

    """
    return matched.expand(config.get("to", r"\1")) if matched is not None and config is not None else value


def compile_re(
    config: tag_publish.configuration.VersionTransform, prefix: str = ""
) -> list[VersionTransform]:
    """
    Compile the from as a regular expression of a dictionary of the config list.

    to be used with convert and match

    Arguments:
        config: The transform config
        prefix: The version prefix

    Return the compiled transform config.

    """
    result = []
    for conf in config:
        new_conf = cast(VersionTransform, dict(conf))

        from_re = conf.get("from", r"(.*)")
        if from_re[0] == "^":
            from_re = from_re[1:]
        if from_re[-1] != "$":
            from_re += "$"
        from_re = f"^{re.escape(prefix)}{from_re}"

        new_conf["from"] = re.compile(from_re)
        result.append(new_conf)
    return result


def match(
    value: str, config: list[VersionTransform]
) -> tuple[Optional[Match[str]], Optional[VersionTransform], str]:
    """
    Get the matched version.

    Arguments:
    ---------
    value: That we want to match with
    config: The result of `compile`

    Returns the re match object, the matched config and the value as a tuple
    On no match it returns None, value

    """
    for conf in config:
        matched = conf["from"].match(value)
        if matched is not None:
            return matched, conf, value
    return None, None, value


def does_match(value: str, config: list[VersionTransform]) -> bool:
    """
    Check if the version match with the config patterns.

    Arguments:
    ---------
    value: That we want to match with
    config: The result of `compile`

    Returns True it it does match else False

    """
    matched, _, _ = match(value, config)
    return matched is not None


def check_response(response: requests.Response, raise_for_status: bool = True) -> Any:
    """
    Check the response and raise an exception if it's not ok.

    Also print the X-Ratelimit- headers to get information about the rate limiting.
    """
    for header in response.headers:
        if header.lower().startswith("x-ratelimit-"):
            print(f"{header}: {response.headers[header]}")
    if raise_for_status:
        response.raise_for_status()


def add_authorization_header(headers: dict[str, str]) -> dict[str, str]:
    """
    Add the Authorization header needed to be authenticated on GitHub.

    Arguments:
        headers: The headers

    Return the headers (to be chained)

    """
    try:
        token = os.environ["GITHUB_TOKEN"].strip()
        headers["Authorization"] = f"Bearer {token}"
        return headers
    except FileNotFoundError:
        return headers


@overload
def download_application(application_name: str, binary_filename: str) -> str: ...


@overload
def download_application(application_name: str) -> None: ...


def download_application(application_name: str, binary_filename: Optional[str] = None) -> Optional[str]:
    """
    Download the application if necessary, with the included version.
    """
    binary_full_filename = (
        os.path.expanduser(os.path.join("~", ".local", "bin", binary_filename)) if binary_filename else None
    )

    if not os.path.exists(binary_full_filename) if binary_full_filename else True:
        applications = applications_download.load_applications(None)
        versions_data = pkgutil.get_data("tag_publish", "versions.yaml")
        assert versions_data is not None
        versions = yaml.safe_load(versions_data)
        applications_download.download_applications(
            applications, {application_name: versions[application_name]}
        )

    return binary_full_filename


def snyk_exec() -> tuple[str, dict[str, str]]:
    """Get the Snyk cli executable path."""
    env = {**os.environ}
    env["FORCE_COLOR"] = "true"

    snyk_bin = download_application("snyk/cli", "snyk")

    if "SNYK_TOKEN" not in env:
        env["SNYK_TOKEN"] = subprocess.run(
            ["gopass", "show", "gs/ci/snyk/token"], check=True, stdout=subprocess.PIPE, encoding="utf-8"
        ).stdout.strip()
    if "SNYK_ORG" in env:
        subprocess.run([snyk_bin, "config", "set", f"org={env['SNYK_ORG']}"], check=True, env=env)

    return snyk_bin, env


class PublishedPayload(TypedDict, total=False):
    """
    The payload to send to the dispatch event.
    """

    type: str
    name: str
    folder: str
    version: str
    tag: str
    repository: str
    version_type: str
    id: int



"""
Automatically generated file from a JSON schema.
"""

from typing import Any, Dict, List, Literal, TypedDict, Union


class Configuration(TypedDict, total=False):
    """
    configuration.

    Tag Publish configuration file
    """

    version: "Version"
    """
    Version.

    The version configurations
    """

    docker: "Docker"
    """
    Docker.

    The configuration used to publish on Docker
    """

    pypi: "Pypi"
    """
    pypi.

    Configuration to publish on pypi
    """

    helm: "Helm"
    """
    helm.

    Configuration to publish Helm charts on GitHub release
    """

    dispatch: List["DispatchConfig"]
    """
    Dispatch.

    default:
      - {}
    """


DISPATCH_CONFIG_DEFAULT: Dict[str, Any] = {}
""" Default value of the field path 'Dispatch item' """


DISPATCH_DEFAULT = [{}]
""" Default value of the field path 'configuration dispatch' """


DISPATCH_EVENT_TYPE_DEFAULT = "published"
""" Default value of the field path 'dispatch config event-type' """


DISPATCH_REPOSITORY_DEFAULT = "camptocamp/argocd-gs-gmf-apps"
""" Default value of the field path 'dispatch config repository' """


DOCKER_AUTO_LOGIN_DEFAULT = False
""" Default value of the field path 'Docker auto_login' """


DOCKER_IMAGE_GROUP_DEFAULT = "default"
""" Default value of the field path 'Docker image group' """


DOCKER_IMAGE_TAGS_DEFAULT = ["{version}"]
""" Default value of the field path 'Docker image tags' """


DOCKER_LATEST_DEFAULT = True
""" Default value of the field path 'Docker latest' """


DOCKER_REPOSITORY_DEFAULT = {
    "github": {"server": "ghcr.io", "versions": ["version_tag", "version_branch", "rebuild"]},
    "dockerhub": {},
}
""" Default value of the field path 'Docker repository' """


DOCKER_REPOSITORY_VERSIONS_DEFAULT = ["version_tag", "version_branch", "rebuild", "feature_branch"]
""" Default value of the field path 'Docker repository versions' """


DOCKER_SNYK_MONITOR_ARGS_DEFAULT = ["--app-vulns"]
""" Default value of the field path 'Docker snyk monitor_args' """


DOCKER_SNYK_TEST_ARGS_DEFAULT = ["--app-vulns", "--severity-threshold=critical"]
""" Default value of the field path 'Docker snyk test_args' """


# | dispatch config.
# |
# | Send a dispatch event to an other repository
# |
# | default:
# |   {}
DispatchConfig = TypedDict(
    "DispatchConfig",
    {
        # | Dispatch repository.
        # |
        # | The repository name to be triggered
        # |
        # | default: camptocamp/argocd-gs-gmf-apps
        "repository": str,
        # | Dispatch event type.
        # |
        # | The event type to be triggered
        # |
        # | default: published
        "event-type": str,
    },
    total=False,
)


class Docker(TypedDict, total=False):
    """
    Docker.

    The configuration used to publish on Docker
    """

    latest: bool
    """
    Docker latest.

    Publish the latest version on tag latest

    default: True
    """

    images: List["DockerImage"]
    """ List of images to be published """

    repository: Dict[str, "DockerRepository"]
    """
    Docker repository.

    The repository where we should publish the images

    default:
      dockerhub: {}
      github:
        server: ghcr.io
        versions:
        - version_tag
        - version_branch
        - rebuild
    """

    auto_login: bool
    """
    Docker auto login.

    Auto login to the GitHub Docker registry

    default: False
    """

    snyk: "_DockerSnyk"
    """ Checks the published images with Snyk """


class DockerImage(TypedDict, total=False):
    """Docker image."""

    group: str
    """
    Docker image group.

    The image is in the group, should be used with the --group option of tag-publish script

    default: default
    """

    name: str
    """ The image name """

    tags: List[str]
    """
    docker image tags.

    The tag name, will be formatted with the version=<the version>, the image with version=latest should be present when we call the tag-publish script

    default:
      - '{version}'
    """


class DockerRepository(TypedDict, total=False):
    """Docker repository."""

    server: str
    """ The server URL """

    versions: List[str]
    """
    Docker repository versions.

    The kind or version that should be published, tag, branch or value of the --version argument of the tag-publish script

    default:
      - version_tag
      - version_branch
      - rebuild
      - feature_branch
    """


HELM_VERSIONS_DEFAULT = ["version_tag"]
""" Default value of the field path 'helm versions' """


class Helm(TypedDict, total=False):
    """
    helm.

    Configuration to publish Helm charts on GitHub release
    """

    folders: List[str]
    """ The folders that will be published """

    versions: List[str]
    """
    helm versions.

    The kind or version that should be published, tag, branch or value of the --version argument of the tag-publish script

    default:
      - version_tag
    """


PIP_PACKAGE_GROUP_DEFAULT = "default"
""" Default value of the field path 'pypi package group' """


PYPI_PACKAGE_FOLDER_DEFAULT = "."
""" Default value of the field path 'pypi package folder' """


PYPI_VERSIONS_DEFAULT = ["version_tag"]
""" Default value of the field path 'pypi versions' """


class Pypi(TypedDict, total=False):
    """
    pypi.

    Configuration to publish on pypi
    """

    packages: List["PypiPackage"]
    """ The configuration of packages that will be published """

    versions: List[str]
    """
    pypi versions.

    The kind or version that should be published, tag, branch or value of the --version argument of the tag-publish script

    default:
      - version_tag
    """


class PypiPackage(TypedDict, total=False):
    """
    pypi package.

    The configuration of package that will be published
    """

    group: str
    """
    pip package group.

    The image is in the group, should be used with the --group option of tag-publish script

    default: default
    """

    folder: str
    """
    pypi package folder.

    The folder of the pypi package

    default: .
    """

    build_command: List[str]
    """ The command used to do the build """


class Version(TypedDict, total=False):
    """
    Version.

    The version configurations
    """

    branch_to_version_re: "VersionTransform"
    """
    Version transform.

    A version transformer definition
    """

    tag_to_version_re: "VersionTransform"
    """
    Version transform.

    A version transformer definition
    """


VersionTransform = List["_VersionTransformItem"]
"""
Version transform.

A version transformer definition
"""


_DOCKER_SNYK_MONITOR_ARGS_ONEOF0_DEFAULT = ["--app-vulns"]
""" Default value of the field path 'Docker Snyk monitor args oneof0' """


_DOCKER_SNYK_MONITOR_ARGS_ONEOF1_DEFAULT = ["--app-vulns"]
""" Default value of the field path 'Docker Snyk monitor args oneof1' """


_DOCKER_SNYK_TEST_ARGS_ONEOF0_DEFAULT = ["--app-vulns", "--severity-threshold=critical"]
""" Default value of the field path 'Docker Snyk test args oneof0' """


_DOCKER_SNYK_TEST_ARGS_ONEOF1_DEFAULT = ["--app-vulns", "--severity-threshold=critical"]
""" Default value of the field path 'Docker Snyk test args oneof1' """


class _DockerSnyk(TypedDict, total=False):
    """Checks the published images with Snyk"""

    monitor_args: Union["_DockerSnykMonitorArgsOneof0", "_DockerSnykMonitorArgsOneof1"]
    """
    Docker Snyk monitor args.

    The arguments to pass to the Snyk container monitor command

    default:
      - --app-vulns

    Aggregation type: oneOf
    """

    test_args: Union["_DockerSnykTestArgsOneof0", "_DockerSnykTestArgsOneof1"]
    """
    Docker Snyk test args.

    The arguments to pass to the Snyk container test command

    default:
      - --app-vulns
      - --severity-threshold=critical

    Aggregation type: oneOf
    """


_DockerSnykMonitorArgsOneof0 = List[str]
"""
default:
  - --app-vulns
"""


_DockerSnykMonitorArgsOneof1 = Literal[False]
"""
default:
  - --app-vulns
"""


_DockerSnykTestArgsOneof0 = List[str]
"""
default:
  - --app-vulns
  - --severity-threshold=critical
"""


_DockerSnykTestArgsOneof1 = Literal[False]
"""
default:
  - --app-vulns
  - --severity-threshold=critical
"""


_VersionTransformItem = TypedDict(
    "_VersionTransformItem",
    {
        # | The from regular expression
        "from": str,
        # | The expand regular expression: https://docs.python.org/3/library/re.html#re.Match.expand
        "to": str,
    },
    total=False,
)



"""
The publishing functions.
"""

import datetime
import glob
import os
import re
import subprocess  # nosec
import sys
from typing import Optional

import ruamel
import tomllib

import tag_publish.configuration


def pip(
    package: tag_publish.configuration.PypiPackage,
    version: str,
    version_type: str,
    publish: bool,
    github: tag_publish.GH,
) -> bool:
    """
    Publish to pypi.

    Arguments:
        version: The version that will be published
        version_type: Describe the kind of release we do: rebuild (specified using --type), version_tag,
                    version_branch, feature_branch, feature_tag (for pull request)
        publish: If False only check the package
        package: The package configuration
        github: The GitHub helper

    """
    folder = package.get("folder", tag_publish.configuration.PYPI_PACKAGE_FOLDER_DEFAULT)
    print(f"::group::{'Publishing' if publish else 'Checking'} '{folder}' to pypi")
    sys.stdout.flush()
    sys.stderr.flush()

    try:
        env = {}
        env["VERSION"] = version
        env["VERSION_TYPE"] = version_type
        default_branch = github.repo.default_branch
        is_master = default_branch == version
        env["IS_MASTER"] = "TRUE" if is_master else "FALSE"

        cwd = os.path.abspath(folder)

        dist = os.path.join(cwd, "dist")
        if not os.path.exists(dist):
            os.mkdir(dist)
        if os.path.exists(os.path.join(cwd, "setup.py")):
            cmd = ["python3", "./setup.py", "egg_info", "--no-date"]
            cmd += (
                ["--tag-build=dev" + datetime.datetime.now().strftime("%Y%m%d%H%M%S")]
                if version_type in ("version_branch", "rebuild")
                else []
            )
            cmd.append("bdist_wheel")
        else:
            if not os.path.exists(dist):
                os.mkdir(dist)
            cmd = ["pip", "wheel", "--no-deps", "--wheel-dir=dist", "."]
            if os.path.exists(os.path.join(cwd, "pyproject.toml")):
                use_poetry = False
                if "build_command" not in package:
                    with open(os.path.join(cwd, "pyproject.toml"), "rb") as project_file:
                        pyproject = tomllib.load(project_file)
                    re_splitter = re.compile(r"[<>=]+")
                    for requirement in pyproject.get("build-system", {}).get("requires", []):
                        requirement_split = re_splitter.split(requirement)
                        if requirement_split[0] in ("poetry", "poetry-core"):
                            use_poetry = True
                            break
                    subprocess.run(
                        ["pip", "install", *pyproject.get("build-system", {}).get("requires", [])], check=True
                    )
                if use_poetry:
                    env_bash = " ".join([f"{key}={value}" for key, value in env.items()])
                    print(f"Run in {cwd}: {env_bash} poetry build")
                    sys.stdout.flush()
                    sys.stderr.flush()
                    subprocess.run(["poetry", "build"], cwd=cwd, env={**os.environ, **env}, check=True)
                    cmd = []
        if cmd:
            cmd = package.get("build_command", cmd)
            subprocess.check_call(cmd, cwd=cwd, env=env)
        cmd = ["twine"]
        cmd += ["upload", "--verbose", "--disable-progress-bar"] if publish else ["check"]
        cmd += glob.glob(os.path.join(cwd, "dist/*.whl"))
        cmd += glob.glob(os.path.join(cwd, "dist/*.tar.gz"))
        subprocess.check_call(cmd)
        print("::endgroup::")
    except subprocess.CalledProcessError as exception:
        print(f"Error: {exception}")
        print("::endgroup::")
        print("::error::With error")
        return False
    return True


def docker(
    config: tag_publish.configuration.DockerRepository,
    name: str,
    image_config: tag_publish.configuration.DockerImage,
    tag_src: str,
    dst_tags: list[str],
    images_full: list[str],
    version_type: str,
    published: Optional[list[tag_publish.PublishedPayload]] = None,
) -> bool:
    """
    Publish to a Docker registry.

    config is like:
        server: # The server fqdn

    image_config is like:
        name: # The image name

    Arguments:
        config: The publishing config
        name: The repository name, just used to print messages
        image_config: The image config
        tag_src: The source tag (usually latest)
        dst_tags: Publish using the provided tags
        images_full: The list of published images (with tag), used to build the dispatch event
        version_type: Describe the kind of release we do: rebuild (specified using --type), version_tag,
                    version_branch, feature_branch, feature_tag (for pull request)
        published: The list of published artifacts to be filled

    """
    print(
        f"::group::Publishing {image_config['name']} to the server {name} "
        f"using the tags {', '.join(dst_tags)}"
    )
    sys.stdout.flush()
    sys.stderr.flush()

    try:
        new_images_full = []
        if "server" in config:
            for tag in dst_tags:
                subprocess.run(
                    [
                        "docker",
                        "tag",
                        f"{image_config['name']}:{tag_src}",
                        f"{config['server']}/{image_config['name']}:{tag}",
                    ],
                    check=True,
                )
                new_images_full.append(f"{config['server']}/{image_config['name']}:{tag}")
                if published is not None:
                    published.append(
                        {
                            "type": "docker",
                            "repository": config["server"],
                            "name": image_config["name"],
                            "tag": tag,
                            "version_type": version_type,
                        }
                    )
        else:
            for tag in dst_tags:
                if tag_src != tag:
                    subprocess.run(
                        [
                            "docker",
                            "tag",
                            f"{image_config['name']}:{tag_src}",
                            f"{image_config['name']}:{tag}",
                        ],
                        check=True,
                    )
                    if published is not None:
                        published.append(
                            {
                                "type": "docker",
                                "repository": "docker.io",
                                "name": image_config["name"],
                                "tag": tag,
                                "version_type": version_type,
                            }
                        )
                new_images_full.append(f"{image_config['name']}:{tag}")

        for image in new_images_full:
            subprocess.run(["docker", "push", image], check=True)
        images_full += new_images_full

        print("::endgroup::")
    except subprocess.CalledProcessError as exception:
        print(f"Error: {exception}")
        print("::endgroup::")
        print("::error::With error")
        return False
    return True


def helm(folder: str, version: str, owner: str, repo: str, commit_sha: str, token: str) -> bool:
    """
    Publish to pypi.

    Arguments:
        folder: The folder to be published
        version: The version that will be published
        owner: The GitHub repository owner
        repo: The GitHub repository name
        commit_sha: The sha of the current commit
        token: The GitHub token

    """
    print(f"::group::Publishing Helm chart from '{folder}' to GitHub release")
    sys.stdout.flush()
    sys.stderr.flush()

    try:
        yaml_ = ruamel.yaml.YAML()
        with open(os.path.join(folder, "Chart.yaml"), encoding="utf-8") as open_file:
            chart = yaml_.load(open_file)
        chart["version"] = version
        with open(os.path.join(folder, "Chart.yaml"), "w", encoding="utf-8") as open_file:
            yaml_.dump(chart, open_file)
        for index, dependency in enumerate(chart.get("dependencies", [])):
            if dependency["repository"].startswith("https://"):
                subprocess.run(["helm", "repo", "add", str(index), dependency["repository"]], check=True)

        subprocess.run(["cr", "package", folder], check=True)
        subprocess.run(
            [
                "cr",
                "upload",
                f"--owner={owner}",
                f"--git-repo={repo}",
                f"--commit={commit_sha}",
                "--release-name-template={{ .Version }}",
                f"--token={token}",
            ],
            check=True,
        )
        if not os.path.exists(".cr-index"):
            os.mkdir(".cr-index")
        subprocess.run(
            [
                "cr",
                "index",
                f"--owner={owner}",
                f"--git-repo={repo}",
                f"--charts-repo=https://{owner}.github.io/{repo}",
                "--push",
                "--release-name-template={{ .Version }}",
                f"--token={token}",
            ],
            check=True,
        )
        print("::endgroup::")
    except subprocess.CalledProcessError as exception:
        print(f"Error: {exception}")
        print("::endgroup::")
        print("::error::With error")
        return False
    return True



"""
Manage OpenID Connect (OIDC) token exchange for external services.

Inspired by
https://github.com/pypa/gh-action-pypi-publish/blob/unstable/v1/oidc-exchange.py
"""

import base64
import json
import os
import sys
from typing import NoReturn

import id as oidc_id
import requests


class _OidcError(Exception):
    pass


def _fatal(message: str) -> NoReturn:
    # HACK: GitHub Actions' annotations don't work across multiple lines naively;
    # translating `\n` into `%0A` (i.e., HTML percent-encoding) is known to work.
    # See: https://github.com/actions/toolkit/issues/193
    message = message.replace("\n", "%0A")
    print(f"::error::Trusted publishing exchange failure: {message}", file=sys.stderr)
    raise _OidcError(message)


def _debug(message: str) -> None:
    print(f"::debug::{message.title()}", file=sys.stderr)


def _render_claims(token: str) -> str:
    _, payload, _ = token.split(".", 2)

    # urlsafe_b64decode needs padding; JWT payloads don't contain any.
    payload += "=" * (4 - (len(payload) % 4))
    claims = json.loads(base64.urlsafe_b64decode(payload))

    return f"""
The claims rendered below are **for debugging purposes only**. You should **not**
use them to configure a trusted publisher unless they already match your expectations.

If a claim is not present in the claim set, then it is rendered as `MISSING`.

* `sub`: `{claims.get('sub', 'MISSING')}`
* `repository`: `{claims.get('repository', 'MISSING')}`
* `repository_owner`: `{claims.get('repository_owner', 'MISSING')}`
* `repository_owner_id`: `{claims.get('repository_owner_id', 'MISSING')}`
* `job_workflow_ref`: `{claims.get('job_workflow_ref', 'MISSING')}`
* `ref`: `{claims.get('ref')}`

See https://docs.pypi.org/trusted-publishers/troubleshooting/ for more help.
"""


def _get_token(hostname: str) -> str:
    # Indices are expected to support `https://{hostname}/_/oidc/audience`,
    # which tells OIDC exchange clients which audience to use.
    audience_resp = requests.get(f"https://{hostname}/_/oidc/audience", timeout=5)
    audience_resp.raise_for_status()

    _debug(f"selected trusted publishing exchange endpoint: https://{hostname}/_/oidc/mint-token")

    try:
        oidc_token = oidc_id.detect_credential(audience=audience_resp.json()["audience"])
    except oidc_id.IdentityError as identity_error:
        _fatal(
            f"""
OpenID Connect token retrieval failed: {identity_error}

This generally indicates a workflow configuration error, such as insufficient
permissions. Make sure that your workflow has `id-token: write` configured
at the job level, e.g.:

```yaml
permissions:
  id-token: write
```

Learn more at https://docs.github.com/en/actions/deployment/security-hardening-your-deployments/about-security-hardening-with-openid-connect#adding-permissions-settings.
"""
        )

    # Now we can do the actual token exchange.
    mint_token_resp = requests.post(
        f"https://{hostname}/_/oidc/mint-token",
        json={"token": oidc_token},
        timeout=5,
    )

    try:
        mint_token_payload = mint_token_resp.json()
    except requests.JSONDecodeError:
        # Token exchange failure normally produces a JSON error response, but
        # we might have hit a server error instead.
        _fatal(
            f"""
Token request failed: the index produced an unexpected
{mint_token_resp.status_code} response.

This strongly suggests a server configuration or downtime issue; wait
a few minutes and try again.

You can monitor PyPI's status here: https://status.python.org/
"""  # noqa: E702
        )

    # On failure, the JSON response includes the list of errors that
    # occurred during minting.
    if not mint_token_resp.ok:
        reasons = "\n".join(
            f'* `{error["code"]}`: {error["description"]}'
            for error in mint_token_payload["errors"]  # noqa: W604
        )

        rendered_claims = _render_claims(oidc_token)

        _fatal(
            f"""
Token request failed: the server refused the request for the following reasons:

{reasons}

This generally indicates a trusted publisher configuration error, but could
also indicate an internal error on GitHub or PyPI's part.

{rendered_claims}
"""
        )

    pypi_token = mint_token_payload.get("token")
    if not isinstance(pypi_token, str):
        _fatal(
            """
Token response error: the index gave us an invalid response.

This strongly suggests a server configuration or downtime issue; wait
a few minutes and try again.
"""
        )

    # Mask the newly minted PyPI token, so that we don't accidentally leak it in logs.
    print(f"::add-mask::{pypi_token}")

    # This final print will be captured by the subshell in `twine-upload.sh`.
    return pypi_token


def pypi_login() -> None:
    """
    Connect to PyPI using OpenID Connect and mint a token for the user.

    See Also
    --------
    - https://docs.github.com/en/actions/security-for-github-actions/security-hardening-your-deployments/about-security-hardening-with-openid-connect
    - https://docs.github.com/en/actions/security-for-github-actions/security-hardening-your-deployments/configuring-openid-connect-in-pypi

    """
    pypirc_filename = os.path.expanduser("~/.pypirc")

    if os.path.exists(pypirc_filename):
        print(f"::info::{pypirc_filename} already exists; consider as already logged in.")  # noqa: E702
        return

    if "ACTIONS_ID_TOKEN_REQUEST_TOKEN" not in os.environ:
        print(
            """::error::Not available, you probably miss the permission `id-token: write`.
              ```
              permissions:
                id-token: write
              ```
              See also: https://docs.github.com/en/actions/security-for-github-actions/security-hardening-your-deployments/about-security-hardening-with-openid-connect"""
        )
        return

    try:
        token = _get_token("pypi.org")
        with open(pypirc_filename, "w", encoding="utf-8") as pypirc_file:
            pypirc_file.write("[pypi]\n")
            pypirc_file.write("repository: https://upload.pypi.org/legacy/\n")
            pypirc_file.write("username: __token__\n")
            pypirc_file.write(f"password: {token}\n")
    except _OidcError:
        # Already visible in logs; no need to re-raise.
        return
